import click

from pymob.utils.store_file import (
    import_package, opt, prepare_casestudy, prepare_scenario)
from pymob.utils import help

@click.command()
@click.option("-c", "--case_study", type=str, default="test_case_study", 
              help=help.case_study)
@click.option("-s", "--scenario", type=str, default="test_scenario", 
              help=help.scenario)
@click.option("-p", "--package", type=str, default="case_studies", 
              help=help.package)
@click.option("-l", "--logging", type=str, default=None)
@click.option("-f", "--logfile", type=str, default=None)
def main(case_study, scenario, package, logging, logfile):
    # TODO: add error messages, in case
    #       - scenario, 
    #       - case_study, 
    #       - package,
    #       - settings.cfg cannot be found
    config = prepare_casestudy(
        case_study=(case_study, scenario), 
        config_file="settings.cfg", 
        pkg_dir=package
    )

    # update parameters from config file if they are specified
    if logging is not None: config.set("case-study", "logging", logging)
    if logfile is not None: config.set("case-study", "logfile", logfile)

    # import package        
    pkg = import_package(package_path=config["case-study"]["package"])
    Simulation = pkg.sim.Simulation

    # create and run simulation
    sim = Simulation(
        config=config
    )
    sim.compute()

    # store and process output
    sim.dump()
    sim.plot()

if __name__ == "__main__":
    main()


import pytest
import numpy as np
from matplotlib import pyplot as plt

from pymob.utils.store_file import prepare_casestudy

def create_simulation(scenario):
    config = prepare_casestudy(
        case_study=("test_case_study", scenario),
        config_file="settings.cfg"
    )
    from case_studies.test_case_study.sim import Simulation
    return Simulation(config=config)

def test_diffrax_exception():
    # with proper scripting API define JAX model here or import from fixtures
    sim = create_simulation("test_scenario")

    # diffrax returns infinity for all computed values after which the solver 
    # breaks due to raching maximum number of steps. 
    # This function calculates the number of inf values
    n_inf = lambda x: (x.results.to_array() == np.inf).values.sum() / len(x.data_variables)

    ub_alpha = 5.0  # alpha values above do not yield reasonable fits for beta = 0.02
    alpha = np.logspace(-2, 3, 20)  # we sample alpha from 0.01-100

    badness = []
    for a in alpha:
        eva = sim.dispatch({"alpha": a, "beta": 0.02})
        eva()

        badness.append(n_inf(eva))

    # the tests make sure that parameters within feasible bounds result in simulation
    # results without inf values and do contain inf values when parameters above
    # feasible bounds are sampled.
    badness_for_feasible_alpha = np.array(badness)[np.where(alpha < ub_alpha)[0]]
    assert sum(badness_for_feasible_alpha) == 0

    badness_for_infeasible_alpha = np.array(badness)[np.where(alpha >= ub_alpha)[0]]
    assert sum(badness_for_infeasible_alpha) > 0


def test_user_defined_probability_model():
    sim = create_simulation("test_scenario")

    sim.config.set("inference.numpyro", "kernel", "nuts")
    sim.config.set("inference.numpyro", "user_defined_probability_model", "parameter_only_model")
    sim.config.set("inference.numpyro", "user_defined_preprocessing", "dummy_preprocessing")

    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    
    posterior_median = sim.inferer.idata.posterior.median(("chain", "draw"))[["beta", "alpha"]]
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_median.to_dataarray().values,
        np.array([0.02, 0.5]),
        rtol=1e-1, atol=1e-3
    )


def test_nuts_kernel():
    sim = create_simulation("test_scenario")

    sim.config.set("inference.numpyro", "kernel", "nuts")
    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    
    posterior_mean = sim.inferer.idata.posterior.mean(("chain", "draw"))[sim.model_parameter_names]
    true_parameters = sim.model_parameter_dict
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_mean.to_dataarray().values,
        np.array(list(true_parameters.values())),
        rtol=1e-2, atol=1e-3
    )

def test_svi_kernel():
    sim = create_simulation("test_scenario")

    sim.config.set("inference.numpyro", "kernel", "svi")
    sim.config.set("inference.numpyro", "svi_iterations", "10000")
    sim.config.set("inference.numpyro", "svi_learning_rate", "0.01")
    # this samples the model with standard normal distributions
    # and rescales them according to the transformations of the specified 
    # parameter distributions to the normal
    sim.config.set("inference.numpyro", "gaussian_base_distribution", "1")

    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    sim.inferer.idata.posterior_predictive

    posterior_mean = sim.inferer.idata.posterior.mean(("chain", "draw"))[sim.model_parameter_names]
    true_parameters = sim.model_parameter_dict
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_mean.to_dataarray().values,
        np.array(list(true_parameters.values())),
        rtol=1e-2, atol=1e-3
    )


    # posterior predictions
    fig, axes = plt.subplots(2,1, sharex=True)
    for data_var, ax in zip(sim.data_variables, axes):
        ax = sim.inferer.plot_posterior_predictions(
            data_variable=data_var, 
            x_dim="time",
            ax=ax
        )


def test_map_kernel():
    sim = create_simulation("test_scenario")

    sim.config.set("inference.numpyro", "kernel", "map")
    sim.config.set("inference.numpyro", "svi_iterations", "2000")
    sim.config.set("inference.numpyro", "svi_learning_rate", "0.01")
    # this samples the model with standard normal distributions
    # and rescales them according to the transformations of the specified 
    # parameter distributions to the normal
    sim.config.set("inference.numpyro", "gaussian_base_distribution", "1")

    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    sim.inferer.idata.posterior_predictive

    posterior_mean = sim.inferer.idata.posterior.mean(("chain", "draw"))[sim.model_parameter_names]
    true_parameters = sim.model_parameter_dict
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_mean.to_dataarray().values,
        np.array(list(true_parameters.values())),
        rtol=1e-2, atol=1e-3
    )



def test_nuts_kernel_replicated():
    pytest.skip()
    # CURRENTLY UNUSABLE SEE https://github.com/flo-schu/pymob/issues/6
    config = prepare_casestudy(
        case_study=("test_case_study", "test_scenario_replicated"),
        config_file="settings.cfg"
    )
    from case_studies.test_case_study.sim import ReplicatedSimulation
    
    sim = ReplicatedSimulation(config=config)
    sim = create_simulation("test_scenario_replicated")

    sim.config.set("inference.numpyro", "kernel", "nuts")
    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    
    posterior_mean = sim.inferer.idata.posterior.mean(("chain", "draw"))[sim.model_parameter_names]
    true_parameters = sim.model_parameter_dict
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_mean.to_dataarray().values,
        np.array(list(true_parameters.values())),
        rtol=1e-2, atol=1e-3
    )

    

def test_sa_kernel():
    sim = create_simulation("test_scenario")

    sim.config.set("inference.numpyro", "kernel", "sa")
    sim.config.set("inference.numpyro", "init_strategy", "init_to_sample")
    sim.config.set("inference.numpyro", "warmup", "2000")
    sim.config.set("inference.numpyro", "draws", "1000")
    sim.config.set("inference.numpyro", "sa_adapt_state_size", "10")

    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    
    posterior_mean = sim.inferer.idata.posterior.mean(("chain", "draw"))[sim.model_parameter_names]
    true_parameters = sim.model_parameter_dict
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_mean.to_dataarray().values,
        np.array(list(true_parameters.values())),
        rtol=1e-2, atol=1e-3
    )



    # posterior predictions
    for data_var in sim.data_variables:
        ax = sim.inferer.plot_posterior_predictions(
            data_variable=data_var, 
            x_dim="time"
        )

if __name__ == "__main__":
    import sys
    import os
    sys.path.append(os.getcwd())
    # test_user_defined_probability_model()


import time
import configparser
import os
import json
import numpy as np
import sys
import xarray as xr
from glob import glob
from importlib import import_module
from pymob.utils.misc import get_host
from pymob.utils.errors import errormsg

def create_fname_date_ver(dirname="./", fname="file", fextension=".txt"):
    ver = 0
    file = dirname + fname + time.strftime("%Y%m%d") + "_" + str(ver) + fextension
    while os.path.isfile(file):
        ver += 1
        file = dirname + fname + time.strftime("%Y%m%d") + "_" + str(ver) + fextension

    return file

def opt(a, b, c):
    """
    tests arguments for their state and returns with priority a, b and then c
    """
    return a if a is not None else b if b is not None else c

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False


def read_config(config_file):
    # get last bit of file name after dot
    ext = config_file.split(".")[-1]

    if ext == "json":
        with open(config_file, "r") as f:
            return json.load(f)

    if ext == "cfg":
        converters={
            'list': lambda x: x if " " not in x else [i.strip() for i in x.split(' ')],
            'strint': lambda x: int(x) if x.isdigit() else x,
            'strfloat': lambda x: float(x) if is_number(x) else x,
            'listfloat': lambda x: x if " " not in x else [float(i.strip()) for i in x.split(' ')]
        }
        config = configparser.ConfigParser(converters=converters)
        _ = config.read(config_file)
        return config

def parse_config_section(section, method: str="strint"):
    """
    method can be one of 'strint' (mix of string and integers) or "list"
    if lists are also in the section
    """
    getter = getattr(section, f"get{method}")
    return {k:getter(k) for k in section.keys()}

def store_sbi_simulations(path, theta, x, simname=None):
    if simname == None:
        simname = time.strftime("%Y%m%d%H%M")
    simloc = os.path.normpath(os.path.join(path, simname))
    os.makedirs(simloc, exist_ok=True)
    np.savetxt(os.path.join(simloc, "theta.txt"), theta.numpy())
    np.savetxt(os.path.join(simloc, "x.txt"), x.numpy())

def read_settings():
    try:
        return read_config("config/settings.json")
    except FileNotFoundError:
        return read_config("config/settings_default.json")

def _insert_username(settings, dir):
    username = settings["username_remote"]
    return dir.replace("_USERNAME_", username)

def prepare_sbi(sbi_config_file):
    config = read_config(sbi_config_file)
    sim_conf = read_config(config["simulation_config"])

    # get the correct file
    settings = read_settings()
    host = get_host(settings["localhosts"])
    output_dir = os.path.join(config[f"output_{host}"])
    
    output = _adapt_output_to_machine(output_dir=output_dir)

    os.makedirs(config["plots"], exist_ok=True)

    return config, sim_conf, output

def unixify_path(path):
    normed_path_os = os.path.normpath(path)
    unix_path = normed_path_os.replace(os.path.sep, "/")
    return unix_path

def scenario_file(file, case_study, scenario, pkg_dir="case_studies"):
    file_path = os.path.join(pkg_dir, case_study, "scenarios", scenario, file)
    if not os.path.exists(file_path):
        cwd = os.getcwd()
        raise FileNotFoundError(
            f"The scenario config file ({file}) was not found in this location: "
            f"'{file_path}'; your working directory is '{cwd}'. "
            "If your working directory is inside the location of your case_study, "
            "try navigating a few levels up the directory tree with "
            "`pkg_dir='..'` or `pkg_dir='../..'`. If your working directory is "
            "outside of the location of your case studies try `pkg_dir='case_studies'`. "
            "If nothing helps, simply provide the absolute path to the parent "
            "directory where your case study is located."
        )
    return file_path

def case_study_output(case_study, scenario, pkg_dir="case_studies"):
    study_name = os.path.basename(case_study)
    output = os.path.normpath(os.path.join(
        pkg_dir, study_name, "results", scenario))
    return pkg_dir, output

def prepare_scenario(case_study, scenario, input_files=[]):
    pkg_dir, output = case_study_output(case_study, scenario)
    
    paths_input_files = []
    for file in input_files:
        fp = scenario_file(file, case_study, scenario)
        paths_input_files.append(fp)

    output_path = reroute_output_to_base(output)

    return paths_input_files, output_path


def reroute_output_to_base(output):
    settings = read_settings()
    return unixify_path(os.path.join(
        settings[f"output_base"], 
        unixify_path(output)))

def prepare_casestudy(
    case_study: tuple[str,str], 
    config_file: str="settings.cfg", 
    pkg_dir="case_studies"
) -> configparser.ConfigParser:
    """Loads the configuration file of the case study by providing the name of 
    the case study directory, the name of the scenario directory, and the name
    of the configuration file. 
    Also adds the case_study directory to `sys.path` so that all modules in the
    root of the case_study directory can be imported (e.g. import sim, 
    import mod, ...)
    
    By default the function assumes that you have the following directory 
    structure:
    
    ```kotlin
    project_directory (working directory)
      ├─ case_studies
      │   ├─ case_study_1
      │   │   ├─ scenarios
      │   │   │   ├─ scenario_1
      │   │   │   │   └─ settings.cfg
      │   │   │   ├─ scenario_2
      │   │   │   │   └─ settings.cfg
      │   │   └─ results
      │   ├─ sim.py
      │   ├─ mod.py
      │   ├─ data.py
      │   ├─ plot.py
      │   └─ ...
      └─ other_directory
    ```
      
    Parameters
    ----------
    case_study : tuple[str,str]
        A tuple of the case_study direcory and the scenario directory
    config_file : str, optional
        The name of the configuration file contained in the scenario directory.
        Default is 'settings.cfg'
    pkg_dir : str, optional
        The name of the folder, where all case studies are located. Can also be
        a relative or an absolute file path. By default "case_studies"

    Returns
    -------
    ConfigParser
        The configuration file of the case-study-scenario combination.

    Example
    -------
    This example demonstrates how to use the prepare_casestudy function:

    ```python
    from my_module import prepare_casestudy

    # Example usage of prepare_casestudy function
    config = prepare_casestudy(("test_case_study", "test_scenario"), "settings.cfg")

    # Now you can access configuration settings
    print(config.get("case-study", "name"))
    print(config.get("case-study", "scenario"))

    # And use the config file to initialize a Simulation object
    from sim import Simulation
    sim = Simulation(config)
    ```
    """
    # read sbi-config, sim-config and settings
    assert len(case_study) == 2, "case_study must be a list with first entry <case_study_directory> and 2nd entry the scenario name"
    config = read_config(scenario_file(config_file, *case_study, pkg_dir=pkg_dir))
    
    config["case-study"]["root"] = os.getcwd()
    config["case-study"]["name"] = case_study[0]
    config["case-study"]["scenario"] = case_study[1]

    # store package dir
    path_case_study = os.path.join(pkg_dir, case_study[0])
    config["case-study"]["package"] = path_case_study

    # append relevant paths to sys
    root = os.path.join(config.get("case-study", "root"), pkg_dir)
    if root not in sys.path:
        sys.path.append(root)
    
    package = os.path.join(config.get("case-study", "root"), config.get("case-study", "package"))
    if package not in sys.path:
        sys.path.append(package)

    return config

def prepare_casestudy_sbi(case_study):
    # read sbi-config, sim-config and settings
    config = read_config(scenario_file("sbi.json", *case_study))
    assert len(case_study) == 2, "case_study must be a list with first entry <case_study_directory> and 2nd entry the scenario name"
    sim_conf = prepare_scenario(*case_study)
    settings = read_settings()

    # determine directory depending of whether working local or remote
    host = get_host(settings["localhosts"])
    output_dir = unixify_path(os.path.join(
        settings[f"{host}_output_base"], 
        unixify_path(sim_conf["simulation"]["output"]).replace(
            "simulation", "sbi")
    ))

    # adapt output to the type of machine being worked on
    output = _adapt_output_to_machine(output_dir=output_dir)
    
    # create plot dir and pass to sbi-config
    plot_dir = unixify_path(sim_conf["simulation"]["output"]).replace(
        "simulation", "plots")
    os.makedirs(plot_dir, exist_ok=True)
    config["plots"] = plot_dir
    
    return config, sim_conf, output

def _adapt_output_to_machine(output_dir):
    settings = read_settings()
    host = get_host(settings["localhosts"])

    if host == "remote":
        output_dir = os.path.normpath(_insert_username(settings, output_dir))

        try:
            os.makedirs(output_dir, exist_ok=True)
        except PermissionError:
            if os.path.isabs(output_dir):
                output_dir_old = output_dir
                output_dir = os.path.abspath(output_dir_old[1:])  # remove leading /
                sys.stdout.write(
                    f"no permission to write to {output_dir_old} using {output_dir}"
                )
                os.makedirs(output_dir, exist_ok=True)
    
    return unixify_path(output_dir)

def import_package(package_path):
    """
    this script handles the import of a case study without the typical 
    __init__.py file. It iterates through all .py files in the root directory
    of the case study (typically: sim, mod, stats, plot, data, prior)
    and imports them with ()
    package     path to package from project dir (e.g. "model/core/daphnia")
    """
    package = os.path.basename(package_path)
    package_dir = os.path.dirname(package_path)
    sys.path.append(package_dir)

    # it is for me only possible to import the specific modules and not 
    # only the package and then refer to the modules
    for mod in glob(os.path.join(package_dir, package, "*.py")):
        module = os.path.basename(mod).split('.')[0]
        try:
            _ = import_module(f"{package}.{module}")
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(errormsg(
                """The module could not be found. This can happen if the name
                of the case-study conflicts with another package. To avoid this
                conflict, give the case study a more unique name. E.g.
                case_study_test. You can confirm, whether another module with that name
                exists, by opening python in the command line and typing the
                name of your case study, if there is no error, it means that the
                package was found in your environment. You can then safely assume
                that the error came from a naming conflict and resolve it as suggested.
                """
            ))

    pkg = import_module(package)
    return pkg

def sequential_filename_iterator(path, sep="_"):
    """
    file storing function that appends a running number to the 
    filename, depending on the number of files in the directory. Note that this
    does only work if files are created sequentially. It will probably produce 
    errors if the files are produced in parallel.
    """
    # split path
    directory, file = os.path.split(path)

    # split filename
    filename, fileext = file.split(".")

    # find number of files in directory
    nfiles = len([
        f for f in os.listdir(directory) 
        if os.path.isfile(os.path.join(directory, f))
    ])
    filename = filename + sep + str(nfiles).zfill(5)

    newpath = os.path.join(directory, filename + "." + fileext)

    return newpath

def serialize(ds, convert_time=True):
    """
    function serializes a dataset. By default converts a time coordinate from
    datetime64[ns] to float[ns]
    """

    if convert_time:
        ds["time"] = ds.time.astype(float)

    return json.dumps(ds.to_dict())

def deserialize(dct, convert_time=True):
    ds = xr.Dataset.from_dict(json.loads(dct))

    if convert_time:
        ds["time"] = ds.time.astype("timedelta64[ns]")

    return ds

def unnest(d, flat, parent_key="", sep="."):
    for child_key, value in d.items():
        if parent_key != "":
            key = f"{parent_key}{sep}{child_key}"
        else:
            key = child_key
            
        if isinstance(value, dict):
            flat = unnest(d=value, flat=flat, parent_key=key)
        else:            
            flat.append((key, value))
    
    return flat


def go_to_case_studies(case_study_dir="case_studies"):
    """A convenience method to find the case studies directory from a directory
    within the case study. And change the working directory respectively"""
    cwd = os.getcwd()
    while not os.path.exists(os.path.join(os.getcwd(), case_study_dir)):
        os.chdir("..")  # navigate one level up

        if os.getcwd() == "/":
            os.chdir(cwd)
            raise NotADirectoryError(
                f"The case study directory {case_study_dir} could not be found"
                "by upwards traversing the directory tree. Please specify "
                "manually the location of the parent of your case_studies folder "
                "by using `os.chdir('path/to/the/parent/of/the/case/studies/folder')`"
            )
        
    print(f"Found case study directory in: {os.getcwd()}")



import os
import glob
import click
import numpy as np
from collections import Counter

from pymob.utils.store_file import prepare_casestudy
from pymob.utils import help

@click.command()
@click.option("-c", "--case_study", type=str, default="core_daphnia", 
              help=help.case_study)
@click.option("-s", "--scenario", type=str, default="expo_control", 
              help=help.scenario)
def main(case_study, scenario):
    # parse sbi config file
    config = prepare_casestudy((case_study, scenario), "sbi.cfg")
    output = config["case-study"].get("output")

    input_dir = os.path.join(output, "sims")
    
    theta_path = os.path.join(output, "theta.txt")
    theta_files = glob.glob(os.path.join(input_dir, "*/theta.txt"))
    theta = np.concatenate(
        [np.loadtxt(f, dtype=np.float32) for f in theta_files])
    np.savetxt(theta_path, theta)

    x_path = os.path.join(output, "x.txt")
    x_files = glob.glob(os.path.join(input_dir, "*/x.txt"))
    x = np.concatenate(
        [np.loadtxt(f, dtype=np.float32) for f in x_files])
    np.savetxt(x_path, x)

    errors = glob.glob(os.path.join(input_dir, "*/errors/*/"))
    errormessages = []
    infos = []
    for i, err in enumerate(errors):
        try:
            with open(os.path.join(err, "log.txt"), "r") as f:
                last_line = f.readlines()[-1]
            
            if "Mass was not conserved." in last_line:
                msg = "mass conservation error"
                info = last_line.split(" ")[3]
            else:
                msg = last_line
                info = ""

            errormessages.append(msg)
            infos.append(info)
            # conf = read_config(os.path.join(err, "config.json"))
        except FileNotFoundError:
            pass

    error_report = Counter(errormessages)
    with open(os.path.join(output, "error_report.txt"), "w") as report:
        for error, number in error_report.items():
            report.write(f"{error}: {number}\n")

if __name__ == "__main__":
    main()



from typing import Literal
import types
import sys
import importlib
import textwrap

def errormsg(msg):
    return textwrap.fill(textwrap.dedent(msg))


INSTALL_MAPPING = {}


def get_version(module: types.ModuleType) -> str:
    version = getattr(module, "__version__", None)

    if version is None:
        raise ImportError(f"Can't determine version for {module.__name__}")
    if module.__name__ == "psycopg2":
        # psycopg2 appends " (dt dec pq3 ext lo64)" to it's version
        version = version.split()[0]
    return version


def import_optional_dependency(
    name: str,
    extra: str = "",
    *,
    errors: Literal["raise", "warn", "ignore"] = "raise",
) -> types.ModuleType | None:
    """
    This is based on pandas.compat._optional.py from 03.04.2024
    https://github.com/pandas-dev/pandas/blob/main/pandas/compat/_optional.py#L88
    The function originally also supports comparison against the version number
    of the installed package.

    Import an optional dependency.

    By default, if a dependency is missing an ImportError with a nice
    message will be raised. If a dependency is present, but too old,
    we raise.

    Parameters
    ----------
    name : str
        The module name.
    extra : str
        Additional text to include in the ImportError message.
    errors : str {'raise', 'warn', 'ignore'}
        What to do when a dependency is not found or its version is too old.

        * raise : Raise an ImportError
        * warn : Only applicable when a module's version is to old.
          Warns that the version is too old and returns None
        * ignore: If the module is not installed, return None, otherwise,
          return the module, even if the version is too old.
          It's expected that users validate the version locally when
          using ``errors="ignore"`` (see. ``io/html.py``)
    min_version : str, default None
        Specify a minimum version that is different from the global pandas
        minimum version required.
    Returns
    -------
    maybe_module : Optional[ModuleType]
        The imported module, when found and the version is correct.
        None is returned when the package is not found and `errors`
        is False, or when the package's version is too old and `errors`
        is ``'warn'`` or ``'ignore'``.
    """
    assert errors in {"warn", "raise", "ignore"}

    package_name = INSTALL_MAPPING.get(name)
    install_name = package_name if package_name is not None else name

    msg = (
        f"Missing optional dependency '{install_name}'. {extra} "
        f"Use pip or conda to install {install_name}."
    )
    try:
        module = importlib.import_module(name)
    except ImportError as err:
        if errors == "raise":
            raise ImportError(msg) from err
        return None

    # Handle submodules: if we have submodule, grab parent module from sys.modules
    parent = name.split(".")[0]
    if parent != name:
        install_name = parent
        module_to_get = sys.modules[install_name]
    else:
        module_to_get = module

    return module


import pytest
import xarray as xr
import numpy as np
from click.testing import CliRunner

from pymob.utils.store_file import prepare_casestudy

def load_test_case_study():
    config = prepare_casestudy(
        case_study=("test_case_study", "test_scenario"),
        config_file="settings.cfg"
    )
    from case_studies.test_case_study.sim import Simulation
    return Simulation(config=config)

def test_scripting_API():
    sim = load_test_case_study()
    evalu = sim.dispatch(theta=sim.model_parameter_dict)
    evalu()

    ds = evalu.results
    ds_ref = xr.load_dataset(f"{sim.data_path}/simulated_data.nc")

    np.testing.assert_allclose(
        (ds - ds_ref).to_array().values,
        0
    )

def test_indexing_simulation():
    pytest.skip()

def test_interactive_mode():
    sim = load_test_case_study()
    sim.interactive()

def test_commandline_API():
    from pymob.simulate import main
    runner = CliRunner()
    
    args = "--case_study=test_case_study --scenario=test_scenario"
    result = runner.invoke(main, args.split(" "))


if __name__ == "__main__":
    import sys
    import os
    sys.path.append(os.getcwd())
    # test_interactive_mode()


# created by Florian Schunck on 01.07.2020
# Project: timepath
# Short description of the feature:
# - add round to nearest integer of X
# 
# ------------------------------------------------------------------------------
# Open tasks:
#
# ------------------------------------------------------------------------------
import numpy as np

def round_to_nearest_multiple_of_y(x, y):
    return np.round(x / y) * y

def take_mean(list):
    if len(list) > 0:
        return np.mean(list)
    else:
        return np.nan



import numpy as np
import xarray as xr
import arviz as az
from matplotlib import pyplot as plt

from pymob.utils.plot_helpers import plot_hist, plot_loghist

def cluster_chains(posterior, deviation="std"):
    assert isinstance(posterior, (xr.DataArray, xr.Dataset))
    chain_means = posterior.mean(dim="draw")
    if deviation == "std":
        chain_dev = posterior.std(dim="draw")
    elif "frac:" in deviation:
        _, frac = deviation.split(":")
        chain_dev = chain_means * float(frac)
    else:
        raise ValueError("Deviation method not implemented.")    

    global cluster_id
    cluster_id = 1
    cluster = [cluster_id] * len(posterior.chain)
    unclustered_chains = posterior.chain.values

    def recurse_clusters(unclustered_chains):
        global cluster_id
        compare = unclustered_chains[0]
        new_cluster = []
        for i in unclustered_chains[1:]:
            a = chain_means.sel(chain=compare) + chain_dev.sel(chain=compare) > chain_means.sel(chain=i)
            b = chain_means.sel(chain=compare) - chain_dev.sel(chain=compare) < chain_means.sel(chain=i)
            isin_dev = (a * b).all()

            if isinstance(isin_dev, xr.Dataset):
                isin_dev = isin_dev.to_array().all()

            if not isin_dev:
                cluster[i] = cluster_id + 1
                new_cluster.append(i)

        cluster_id += 1
        if len(new_cluster) == 0:
            return

        recurse_clusters(new_cluster)
    
    recurse_clusters(unclustered_chains)

    return cluster


def rename_extra_dims(df, extra_dim_suffix="_dim_0", new_dim="new_dim", new_coords=None):
    # TODO: COuld be used for numypro backend for fixing posterior indexes
    df_ = df.copy()
    data_vars = list(df_.data_vars.keys())

    # swap dimension names for all dims that have the suffix 
    new_dims = {}
    for dv in data_vars:
        old_dim = f"{dv}{extra_dim_suffix}"
        if df_.dims[old_dim] == 1:
            df_[dv] = df_[dv].squeeze(old_dim)
        else:
            new_dims.update({old_dim: new_dim})

    df_ = df_.swap_dims(new_dims)

    # assign coords to new dimension
    df_ = df_.assign_coords({new_dim: new_coords})
    
    # drop renamed coords
    df_ = df_.drop([f"{dv}{extra_dim_suffix}" for dv in data_vars])

    return df_



# plot loghist
def plot_posterior_samples(posterior, col_dim=None, log=True, hist_kwargs={}):
    if log:
        hist = plot_loghist
    else:
        hist = plot_hist

    parameters = list(posterior.data_vars.keys())
    samples = posterior.stack(sample=("chain", "draw"))

    fig = plt.figure(figsize=(5, len(parameters)*2))
    fig.subplots_adjust(right=.95, top=.95, hspace=.25)

    gs = fig.add_gridspec(len(parameters), 1)

    for i, key in enumerate(parameters):
        postpar = samples[key]
        if col_dim in postpar.dims:
            col_coords = postpar[col_dim]
            gs_par = gs[i, 0].subgridspec(1, len(col_coords))
            axes = gs_par.subplots()

            for ax, coord in zip(axes, col_coords):
                hist(
                    x=postpar.sel({col_dim: coord}), 
                    name=f"${key}$ {str(coord.values)}",
                    ax=ax,
                    **hist_kwargs
                )

        else:
            gs_par = gs[i, 0].subgridspec(1, 1)
            ax = gs_par.subplots()

            hist(
                x=postpar, 
                name=f"${key}$",
                ax=ax,
                **hist_kwargs
            )


    return fig


def bic(idata: az.InferenceData):
    """calculate the BIC for az.InferenceData. The function will average over
    all samples from the markov chain
    """
    log_likelihood = idata.log_likelihood.mean(("chain", "draw")).sum().to_array().sum()
    k = idata.posterior.mean(("chain", "draw")).count().to_array().sum()

    vars = [i.split("_")[0] for i in list(idata.log_likelihood.data_vars.keys())]
    n = 0
    for v in vars:
        if v in idata.observed_data:
            n += (~idata.observed_data[v].isnull()).sum()
        elif v + "_obs" in idata.observed_data:
            n += (~idata.observed_data[v + "_obs"].isnull()).sum()
        else:
            raise IndexError(f"Variable {v} or {v+'_obs'} not in idata")

    # n = (~idata.observed_data[vars].isnull()).sum().to_array().sum()

    bic = float(k * np.log(n) - 2 * log_likelihood)
    msg = str(
        "Bayesian Information Criterion (BIC)"
        "\n===================================="
        f"\nParameters: {int(k)}"
        f"\nData points: {int(n)}"
        f"\nLog-likelihood: {float(log_likelihood)}"
        f"\nBIC: {bic}"
    )

    return msg, bic


def add_cluster_coordinates(idata, deviation="std"):
    cluster = cluster_chains(idata.posterior, deviation=deviation)
    idata = idata.assign_coords(cluster=("chain", cluster))
    return idata


from typing import Literal
import inspect
import xarray as xr

class SolverBase:
    """
    The idea of creating a solver as a class is that it is easier
    to pass on important arguments of the simulation relevant to the 
    Solver. Therefore a solver can access all attributes of an Evaluator
    """
    def __call__(self, **kwargs):
        return self.solve(**kwargs)
    
    @staticmethod
    def solve():
        raise NotImplementedError("Solver must implement a solve method.")


def mappar(func, parameters, exclude=[]):
    func_signature = inspect.signature(func).parameters.keys()
    model_param_signature = [p for p in func_signature if p not in exclude]
    model_args = [parameters.get(k) for k in model_param_signature]

    return tuple(model_args)


def create_interpolation(
        x_in: xr.Dataset, 
        y: str, 
        x: str="time", 
        factor: float=1e-4, 
        interpolation: Literal["fill-forward", "linear"] = "fill-forward",
    ) -> xr.Dataset:
    """Make the interpolation safe by adding a coordinate just before each 
    x-value (except the first vaue). The distance between the new and the next
    point are calculated as a fraction of the previous distance between
    neighboring points. The corresponding y-values are first set to NaN and then
    interpolated based on the interpolation method.

    Parameters
    ----------
    x_in : xr.Dataset
        The input dataset which contains a coordinate (x) and a data variable
        (y)
    x : str, optional
        The name of the x coordinate, by default "time"
    factor : float, optional
        The distance between the newly added points and the following existing
        points on the x-scale, by default 1e-4
    interpolation : Literal["fill-forward", "linear"], optional
        The interpolation method. In addition to 'fill-forward' and 'linear',
        any method give in `xr.interpolate_na` can be chosen, by default
        "fill-forward"

    Returns
    -------
    xr.Dataset
        The interpolated dataset
    """
    xs = x_in.coords[x]

    # calculate x values that are located just a little bit smaller than the xs
    # where "just a little bit" is defined by the distance to the previous x
    # and a factor. This way the scale of the observations should not matter
    # and even very differently sized x-steps should be interpolated correctly
    fraction_before_xs = (
        xs.isel({x:range(1, len(xs))}).values
        - xs.diff(dim=x) * factor
    )

    # create a sorted time vector
    xs = sorted([*fraction_before_xs.values, *xs.values])

    # add new time indices with NaN values 
    x_in_reindexed = x_in.reindex({x:xs})

    if interpolation == "fill-forward":
        # then fill nan values with the previous value (forward-fill)
        x_in_interpolated = x_in_reindexed.ffill(dim=x, limit=1)

    else:
        x_in_interpolated = x_in_reindexed.interpolate_na(dim=x, method="linear")

    return x_in_interpolated


import json

import numpy as np
from sklearn.preprocessing import MinMaxScaler
from matplotlib import pyplot as plt

from pymob.simulation import SimulationBase
from pymob.utils.errors import import_optional_dependency

extra = "'pymoo' dependencies can be installed with pip install pymob[pymoo]"
pymoo = import_optional_dependency("pymoo", errors="warn", extra=extra)
if pymoo is not None:
    import pathos.multiprocessing as mp
    from pymoo.algorithms.moo.unsga3 import UNSGA3
    from pymoo.core.problem import ElementwiseProblem, Problem
    from pymoo.util.ref_dirs import get_reference_directions
    from pymoo.termination.default import DefaultMultiObjectiveTermination


class PymooBackend:
    def __init__(
        self, 
        simulation: SimulationBase,
    ):
        self.simulation = simulation
        self.pool: mp.Pool = None
        self.distance_function = self.distance_function_parser()
        self.transform = self.variable_parser()
        self.problem = OptimizationProblem(backend=self)

    @property
    def population_size(self):
        return self.simulation.config.getint("inference.pymoo", "population_size", fallback=100)
    
    @property
    def seed(self):
        return self.simulation.config.getint("inference.pymoo", "seed", fallback=1)
        
    @property
    def max_nr_populations(self):
        return self.simulation.config.getint("inference.pymoo", "max_nr_populations", fallback=50)
    
    @property
    def xtol(self):
        return self.simulation.config.getfloat("inference.pymoo", "xtol", fallback=0.0005)
    
    @property
    def ftol(self):
        return self.simulation.config.getfloat("inference.pymoo", "ftol", fallback=0.005)
    
    @property
    def cvtol(self):
        return self.simulation.config.getfloat("inference.pymoo", "cvtol", fallback=1e-8)
    
    @property
    def verbose(self):
        return self.simulation.config.get("inference.pymoo", "verbose", fallback=True)
    
    def distance_function_parser(self):
        def f(x):
            Y = self.simulation.evaluate(theta=x)
            obj_name, obj_value = self.simulation.objective_function(results=Y)
            return obj_value
        return f

    def variable_mapper(self, x):
        names = self.simulation.model_parameter_names
        return {n:x_i for n, x_i in zip(names, x)}

    def variable_parser(self):
        variables = self.simulation.free_model_parameters

        bounds = []
        names = []
        for v in variables:
            bounds.append([v.min, v.max])
            names.append(v.name)

        bounds = np.array(bounds).T
        scaler = MinMaxScaler().fit(bounds)

        # check that parameter values and names match, so variable_mapper 
        # can be safely called
        assert names == self.simulation.model_parameter_names

        def transform(X_scaled):
            X = scaler.inverse_transform(X_scaled)
            return map(self.variable_mapper, X)

        return transform
    
    def store_results(self):
        res = self.result

        params = list(self.transform(X_scaled=np.array(res.X, ndmin=2)) )[0]

        res_dict = {
            "f": res.f,
            "cv": res.cv,
            "X": params
        }
        
        file = f"{self.simulation.output_path}/pymoo_params.json"
        with open(file, "w") as fp:
            json.dump(res_dict, fp, indent=4)

        print(f"written results to {file}")

    def run(self):
        """Implements the parallelization in pymoo"""
        n_cores = self.simulation.n_cores
        print(f"Using {n_cores} CPU cores")

        if n_cores == 1:
            self.optimize()
        elif n_cores > 1:
            with mp.ProcessingPool(n_cores) as pool:
                self.pool = pool  # assign pool so it can be accessed by _evaluate
                self.optimize()


    def optimize(self):

        reference_directions = get_reference_directions(
            name="energy", 
            n_dim=self.simulation.n_objectives, 
            n_points=self.population_size,
            seed=self.seed + 1
        )

        algorithm = UNSGA3(
            ref_dirs=reference_directions,
            pop_size=self.population_size
        )

        termination = DefaultMultiObjectiveTermination(
            xtol=self.xtol,
            cvtol=self.cvtol,
            ftol=self.ftol,
            n_max_gen=self.max_nr_populations,
        )

        # prepare the algorithm to solve the specific problem (same arguments as for the minimize function)
        algorithm.setup(
            problem=self.problem, 
            termination=termination,
            seed=self.seed, 
            verbose=True,
        )


        # until the algorithm has no terminated
        while algorithm.has_next():

            # ask the algorithm for the next solution to be evaluated
            pop = algorithm.ask()

            # evaluate the individuals using the algorithm's evaluator (necessary to count evaluations for termination)
            algorithm.evaluator.eval(self.problem, pop)

            # returned the evaluated individuals which have been evaluated or even modified
            algorithm.tell(infills=pop)

            # self.post_processing(pop=pop)

        # obtain the result objective from the algorithm
        res = algorithm.result()

        # save the results
        self.result = res
        self.store_results()

    def post_processing(self, pop):
        F = pop.get("F")
        X = pop.get("X")

        f_min = F.min()
        x_min = X[np.where(F[:,0] == f_min)]
        # print(
        #     f"Generation {gen}: f_min={f_min}, x_min={x_min}",
        #     flush=True
        # )

    def load_results(self):
        with open(f"{self.simulation.output_path}/pymoo_params.json", "r") as fp:
            self.result = json.load(fp)
            
    def plot_predictions(
            self, data_variable: str, x_dim: str, ax=None, subset={}, 
            upscale_x=True
        ):
        obs = self.simulation.observations.sel(subset)
        x_old = self.simulation.coordinates[x_dim].copy()
        
        nan_obs = obs[data_variable].isnull().all(dim=x_dim)

        if x_dim is not None:    
            if upscale_x:
                self.simulation.coordinates[x_dim] = np.linspace(
                    x_old.min(),
                    x_old.max(),
                    1000
                )
            else:
                raise NotImplementedError("x_new must be a 1D np.ndarray")
            
        Y = self.simulation.evaluate(self.result["X"])
        results = self.simulation.results_to_df(results=Y)

        if ax is None:
            ax = plt.subplot(111)
        
        ax.plot(
            results[x_dim].values, 
            results[data_variable].values.T[:, ~nan_obs], 
            color="black", lw=.8
        )

        ax.plot(
            obs[x_dim].values, 
            obs[data_variable].values[:, ~nan_obs], 
            marker="o", ls="", ms=3
        )
        
        ax.set_ylabel(data_variable)
        ax.set_xlabel(x_dim)

        # restore old coordinates
        self.simulation.coordinates[x_dim] = x_old

        return ax

if pymoo is not None:
    class OptimizationProblem(Problem):
        def __init__(self, backend: PymooBackend, **kwargs):
            self.backend = backend
            self.simulation = backend.simulation
            
            super().__init__(
                n_var=self.simulation.n_free_parameters, 
                n_obj=self.simulation.n_objectives, 
                n_ieq_constr=0, 
                xl=0.0, 
                xu=1.0,
                elementwise_evaluation=False,
                **kwargs
            )

        def _evaluate(self, X, out, *args, **kwargs):
            X_original_scale = self.backend.transform(X_scaled=X)       

            if self.backend.pool is None:
                F = list(map(self.backend.distance_function, X_original_scale))
            else:
                F = self.backend.pool.map(self.backend.distance_function, X_original_scale)

            out["F"] = np.array(F)


import os
import re
from functools import lru_cache
import tempfile
import inspect

import numpy as np
import xarray as xr
import arviz as az
from matplotlib import pyplot as plt

from pymob.simulation import SimulationBase
from pymob.utils.store_file import is_number
from pymob.utils.errors import import_optional_dependency

extra = "'pyabc' dependencies can be installed with pip install pymob[pyabc]"
pyabc = import_optional_dependency("pyabc", errors="warn", extra=extra)
if pyabc is not None:
    import pyabc
    from pathos import multiprocessing as mp

class PyabcBackend:
    def __init__(
        self, 
        simulation: SimulationBase
    ):
        self.parameter_map = {}
        
        self.simulation = simulation
        self.evaluator = self.model_parser()
        self.prior = self.prior_parser(simulation.free_model_parameters)
        self.distance_function = self.distance_function_parser()
        self.observations = simulation.observations

        self.abc = None
        self.history = None
        self.posterior = None


    @property
    def extra_vars(self):
        extra = self.simulation.config.getlist( 
            "inference", "extra_vars", fallback=[]
        )
        return extra if isinstance(extra, list) else [extra]
    
    
    @property
    def sampler(self):
        return self.simulation.config.get("inference.pyabc", "sampler")
    
    @property
    def population_size(self):
        return self.simulation.config.getint("inference.pyabc", "population_size")
    
    @property
    def minimum_epsilon(self):
        return self.simulation.config.getfloat("inference.pyabc", "minimum_epsilon")
    
    @property
    def min_eps_diff(self):
        return self.simulation.config.getfloat("inference.pyabc", "min_eps_diff")
    
    @property
    def max_nr_populations(self):
        return self.simulation.config.getint("inference.pyabc", "max_nr_populations")
    
    @property
    def database(self):
        tmp = tempfile.gettempdir()
        dbp = self.simulation.config.get(
            "inference.pyabc", "database_path", fallback=f"{tmp}/pyabc.db"
        )
        if os.path.isabs(dbp):
            return dbp
        else:
            return os.path.join(self.simulation.output_path, dbp)
    
    @property
    def redis_password(self):
        return self.simulation.config.get("inference.pyabc.redis", "password", "nopassword")
    
    @property
    def redis_port(self):
        return self.simulation.config.getint("inference.pyabc.redis", "port", 1111)
    
    @property
    def history_id(self):
        return self.simulation.config.getint(
            "inference.pyabc", "eval.history_id", fallback=-1
        )

    @property
    def model_id(self):
        return self.simulation.config.getint(
            "inference.pyabc", "eval.model_id", fallback=0
        )
    
    @property
    def n_predictions(self):
        return self.simulation.config.getint(
            "inference.pyabc", "eval.n_predictions", fallback=50
        )
    
    @staticmethod
    def param_to_prior(par):
        parname = par.name
        distribution, cluttered_arguments = par.prior.split("(", 1)
        param_strings = cluttered_arguments.split(")", 1)[0].split(",")
        params = {}
        arraypattern = r"\[(\d+(\.\d+)?(\s+\d+(\.\d+)?)*|\s*)\]"
        
        for parstr in param_strings:

            key, expression = parstr.split("=")
            
            # check if val is a number
            if is_number(expression):
                value = float(expression)
            
            # check if val is an array
            elif re.fullmatch(arraypattern, expression):
                expression = expression.removeprefix("[").removesuffix("]")
                value = np.array([float(v) for v in expression.split(" ")])

            else:
                raise NotImplementedError(
                    f"Prior format {expression} is not implemented. "
                    "Use e.g."
                    "\n- normal(loc=1.0,scale=0.5) for 1-dimensional priors, or"
                    "\n- normal(loc=[1.0 2.4 1],scale=0.5) for n-dimensional priors." 
                )

            params.update({key:value})

        return parname, distribution, params

    def prior_parser(self, free_model_parameters: list):
        prior_dict = {}
        for mp in free_model_parameters:
            name, distribution, params = self.param_to_prior(par=mp)
            pmap, dist_map = self.array_param_to_1d(name, distribution, params)

            self.parameter_map.update(pmap)
            for subpar_name, subpar_dist in dist_map.items():
                prior = pyabc.RV(subpar_dist["dist"], **subpar_dist["params"])
                prior_dict.update({subpar_name: prior})

        return pyabc.Distribution(**prior_dict)

    @staticmethod
    def map_parameters(theta, parameter_map):
        theta_mapped = {}
        for par_name, subpar_name in parameter_map.items():
            if isinstance(subpar_name, list):
                subparams = np.array([theta[p] for p in subpar_name])
            else:
                subparams = theta[subpar_name]
            
            theta_mapped.update({par_name: subparams})
        return theta_mapped

    @staticmethod
    def array_param_to_1d(name, distribution, dist_param_dict):
        # return a distribution if all parameters of the distribution are floats
        # and map with name

        if np.all([isinstance(v, float) for _, v in dist_param_dict.items()]):
            return {name:name}, {name: {"dist": distribution, "params": dist_param_dict}}
        else:
            args_as_arrays = {k:np.array(v, ndmin=1) for k, v in dist_param_dict.items()}
            broadcasted_args = np.broadcast_arrays(*tuple(args_as_arrays.values()))
            # args = {k: v for k, v in zip(dist_param_dict.keys(), broadcasted_args)}

            param_map = {name: []}
            dist_map = {}
            for i, args in enumerate(zip(*broadcasted_args)):
                p_subname = f"{name}_{i}"
                param_map[name].append(p_subname)
                dist_kwargs = {
                    "dist": distribution,
                    "params": {k:v for k, v in zip(dist_param_dict.keys(), args)}
                }
                dist_map.update({p_subname: dist_kwargs})

            return param_map, dist_map

    
    @property
    def plot_function(self):
        return self.simulation.config.get(
            "inference.pyabc", "plot_function", 
            fallback=None
        )

    def plot(self):
        plot_func = getattr(self.simulation, self.plot_function)
        plot_func()
        

    def model_parser(self):
        extra_kwargs = {}
        for extra in self.extra_vars:
            extra_kwargs.update({extra: self.simulation.observations[extra].values})

        obj_func_signature = inspect.signature(self.simulation.objective_function)
        obj_func_params = list(obj_func_signature.parameters.keys())
    
        def model(theta):
            theta_mapped = self.map_parameters(theta, self.parameter_map)
            evaluator = self.simulation.dispatch(theta=theta_mapped, **extra_kwargs)
            evaluator(seed=np.random.randint(1e6))
            res = {k: np.array(v) for k, v in evaluator.Y.items()}
            res.update({p: theta_mapped[p] for p in obj_func_params if p in theta_mapped})
            return res
        return model
    
    def distance_function_parser(self):
        obj_func_signature = inspect.signature(self.simulation.objective_function)
        obj_func_params = list(obj_func_signature.parameters.keys())
            
        def distance_function(x, x0):
            Y = {k: v for k, v in x.items() if k in self.simulation.data_variables}
            
            theta_obj_func = {p: x[p] for p in obj_func_params if p in x}
            obj_name, obj_value = self.simulation.objective_function(
                results=Y, **theta_obj_func 
            )
            return obj_value
        
        return distance_function

    def run(self):
        n_cores = self.simulation.n_cores
        print(f"Using {n_cores} CPU cores", flush=True)

        # before launch server in bash with `redis-server --port 1803`
        if self.sampler.lower() == "RedisEvalParallelSampler".lower():
            abc_sampler = pyabc.sampler.RedisEvalParallelSampler(
                host="localhost", 
                password=self.redis_password, 
                port=self.redis_port
            )

        elif self.sampler.lower() == "SingleCoreSampler".lower():
            abc_sampler = pyabc.sampler.SingleCoreSampler()

        elif self.sampler.lower() == "MulticoreParticleParallelSampler".lower():
            abc_sampler = pyabc.sampler.MulticoreParticleParallelSampler(
                n_procs=n_cores
            )

        elif self.sampler.lower() == "MulticoreEvalParallelSampler".lower():
            abc_sampler = pyabc.sampler.MulticoreEvalParallelSampler(
                n_procs=n_cores
            )

        else:
            raise NotImplementedError(
                "Sampler is not implemented. Choose one of: 'RedisEvalParallelSampler', " +
                "'SingleCoreSampler'"
            )


        self.abc = pyabc.ABCSMC(
            models=self.evaluator, 
            parameter_priors=self.prior, 
            distance_function=self.distance_function, 
            sampler=abc_sampler,
            population_size=self.population_size
        )

        self.history = self.abc.new("sqlite:///" + self.database)

        self.abc.run(
            minimum_epsilon=self.minimum_epsilon,
            min_eps_diff=self.min_eps_diff,
            max_nr_populations=self.max_nr_populations
        )


    def load_results(self):
        if self.history is None:
            self.history = pyabc.History(f"sqlite:///" + self.database)
        
        # set history id
        db_id = self.history_id
        self.history.id = self.history._find_latest_id() if db_id == -1 else db_id
        
        mod_id = self.model_id
        samples, w = self.history.get_distribution(m=mod_id, t=self.history.max_t)
        
        # re-sort parameters based on prior order
        samples = samples[self.prior.keys()]

        posterior = xr.DataArray(
            samples.values.reshape((1,*samples.values.shape)),
            coords=dict(
                chain=[1],
                draw=range(len(samples)),
                parameter=list(samples.columns)
            )
        )
        
        idata = az.from_dict(
            posterior={key: col.values for key, col in samples.items()},
            dims=self.posterior_data_structure,
            coords=self.posterior_coordinates
        )
        self.idata = idata
        # posterior
        self.posterior = Posterior(posterior)

    @property
    def posterior_data_structure(self):
        data_structure = self.simulation.data_structure.copy()
        data_structure_loglik = {f"{dv}_obs": dims for dv, dims in data_structure.items()}
        data_structure.update(data_structure_loglik)
        return data_structure

    @property
    def posterior_coordinates(self):
        posterior_coords = self.simulation.coordinates.copy()
        posterior_coords.update({
            "draw": list(range(self.population_size)), 
            "chain": [0]
        })
        return posterior_coords

    def plot_chains(self):

        distributions = []
        for t in range(self.history.max_t + 1):
        # for t in range(2):
            print(f"iteration: {t}", end="\r")
            par_values, _ = self.history.get_distribution(m=0, t=t)
            post = par_values.to_xarray().to_array()
            post["id"] = range(len(post.id))
            post = post.assign_coords(iteration=t)
            distributions.append(post)

        trajectory = xr.concat(distributions, dim="iteration")
        
        parameters = trajectory.coords["variable"].values
        nit = len(trajectory.iteration)
        color="tab:blue"
        Np = len(parameters)
        fig, axes = plt.subplots(nrows=Np, ncols=2, figsize=(10, 2*Np),
            gridspec_kw=dict(width_ratios=[1, 1]))
        
        if len(axes.shape) == 1:
            axes = axes.reshape((Np, 2))
        
        for i, par in enumerate(parameters):
            ax = axes[i, 0]

            ax.plot(trajectory.iteration, trajectory.sel(variable=par), 
                color=color,
                # s=5,
                alpha=.01)
                
            # takes the last 
            ax2 = axes[i, 1]
            ax2.hist(trajectory.sel(variable=par, iteration=nit-1), 
                    alpha=.85, color=color)
            ax2.set_ylabel(f"{par.split('.')[-1].replace('_', ' ')}")
            
            ax.set_xlabel("iteration")
            ax.set_yscale("log")
            ax.set_ylabel(f"{par.split('.')[-1].replace('_', ' ')}")

        fig.subplots_adjust(wspace=.3, hspace=.4)

        return fig

    def store_results(self):
        """results are stored by default in database"""
        pass
   
    @lru_cache
    def posterior_predictions(self, n=50, seed=1):
        rng = np.random.default_rng(seed)
        post = self.posterior
        total_samples = post.samples.shape[1]

        # draw samples from posterior
        posterior_samples = rng.choice(post.samples.draw, size=n, replace=False)

        def predict(posterior_sample_id):
            params = post.draw(i=posterior_sample_id)
            evaluator = self.simulation.dispatch(params.to_dict())
            evaluator()
            res = evaluator.results
            res = res.assign_coords({"draw": posterior_sample_id, "chain": 1})
            res["params"] = params.samples
            res = res.expand_dims(("chain", "draw"))
            return res
        
        print(f"Using {self.simulation.n_cores} CPUs")
        if self.simulation.n_cores == 1:
            results = list(map(predict, posterior_samples))
        else:
            with mp.ProcessingPool(self.simulation.n_cores) as pool:        
                results = pool.map(predict, posterior_samples)
            
        return xr.combine_by_coords(results)

    def plot_predictions(
            self, data_variable: str, x_dim: str, ax=None, subset={}
        ):
        obs = self.simulation.observations.sel(subset)
        
        post_pred = self.posterior_predictions(
            n=self.n_predictions, 
            # seed only controls the parameters samples drawn from posterior
            seed=self.simulation.seed
        ).sel(subset)

        hdi = az.hdi(post_pred, .95)

        if ax is None:
            ax = plt.subplot(111)
        
        y_mean = post_pred[data_variable].mean(dim=("chain", "draw"))
        ax.plot(
            post_pred[x_dim].values, y_mean.values, 
            color="black", lw=.8
        )

        ax.fill_between(
            post_pred[x_dim].values, *hdi[data_variable].values.T, 
            alpha=.5, color="grey"
        )

        ax.plot(
            obs[x_dim].values, obs[data_variable].values, 
            marker="o", ls="", ms=3
        )
        
        ax.set_ylabel(data_variable)
        ax.set_xlabel(x_dim)

        return ax

class Posterior:
    def __init__(self, samples):
        self.samples = samples

    def __repr__(self):
        return str(self.samples)

    def to_dict(self):
        theta = self.samples
        return {par:float(theta.sel(parameter=par)) 
            for par in theta.parameter.values}

    def draw(self, i):
        return Posterior(self.samples.sel(draw=i, chain=1))

    def mean(self):
        return Posterior(self.samples.mean(dim=("chain", "draw")))





import xarray as xr

def cluster_chains(posterior):
    assert isinstance(posterior, xr.DataArray)
    chain_means = posterior.mean(dim="draw")
    chain_sd = posterior.std(dim="draw")
    

    global cluster_id
    cluster_id = 1
    cluster = [cluster_id] * len(posterior.chain)
    unclustered_chains = posterior.chain.values

    def recurse_clusters(unclustered_chains):
        global cluster_id
        compare = unclustered_chains[0]
        new_cluster = []
        for i in unclustered_chains[1:]:
            a = chain_means.sel(chain=compare) + chain_sd.sel(chain=compare) > chain_means.sel(chain=i)
            b = chain_means.sel(chain=compare) - chain_sd.sel(chain=compare) < chain_means.sel(chain=i)

            if not all(a * b):
                cluster[i] = cluster_id + 1
                new_cluster.append(i)

        cluster_id += 1
        if len(new_cluster) == 0:
            return

        recurse_clusters(new_cluster)
    
    recurse_clusters(unclustered_chains)

    return cluster


from functools import partial, lru_cache
import glob
import re
import warnings
from typing import Tuple, Dict, Union, Optional, Callable

from tqdm import tqdm
import numpy as np
import xarray as xr
import arviz as az
from matplotlib import pyplot as plt
import sympy

from pymob.simulation import SimulationBase
from pymob.inference.analysis import (
    cluster_chains, rename_extra_dims, plot_posterior_samples,
    add_cluster_coordinates
)

import numpyro
from numpyro import distributions as dist
from numpyro.infer import Predictive
from numpyro.distributions import Normal, transforms, TransformedDistribution
from numpyro import infer
import jax
import jax.numpy as jnp
import sympy2jax

sympy2jax_extra_funcs = {
    sympy.Array: jnp.array,
    sympy.Tuple: tuple,
}

def LogNormalTrans(loc, scale):
    return TransformedDistribution(
        base_distribution=Normal(0,1), 
        transforms=[
            transforms.AffineTransform(loc=jnp.log(loc), scale=scale), 
            exp()
        ])

def HalfNormalTrans(scale):
    return TransformedDistribution(
        base_distribution=Normal(0,1), 
        transforms=[
            transforms.AffineTransform(loc=0, scale=scale), 
            transforms.AbsTransform()
        ])


def transform(transforms, x):
    for part in transforms:
        x = part(x)
    return x

def inv_transform(transforms, y):
    for part in transforms[::-1]:
        y = part.inv(y)
    return y


def catch_patterns(expression_str):
    # tries to match array notation [0 1 2]
    pattern = r"\[(\d+(\.\d+)?(\s+\d+(\.\d+)?)*|\s*)\]"
    if re.fullmatch(pattern, expression_str) is not None:
        expression_str = expression_str.replace(" ", ",") \
            .removeprefix("[").removesuffix("]")
        return f"stack({expression_str})"

    return expression_str

def generate_transform(expression_str):
    # check for parentheses in expression
    
    expression_str = catch_patterns(expression_str)

    # Parse the expression without knowing the symbol names in advance
    parsed_expression = sympy.sympify(expression_str, evaluate=False)
    free_symbols = tuple(parsed_expression.free_symbols)

    # Transform expression to jax expression
    func = sympy2jax.SymbolicModule(
        parsed_expression, 
        extra_funcs=sympy2jax_extra_funcs, 
        make_array=True
    )

    return {"transform": func, "args": [str(s) for s in free_symbols]}

exp = transforms.ExpTransform
sigmoid = transforms.SigmoidTransform
C = transforms.ComposeTransform

class NumpyroBackend:
    def __init__(
        self, 
        simulation: SimulationBase
    ):
        """Initializes the NumpyroBackend with a Simulation object.

        Parameters
        ----------
        simulation : SimulationBase
            An initialized simulation.
        """
        self.EPS = 1e-8
        self.distribution_map = {
            "lognorm": (LogNormalTrans, {"scale": "loc", "s": "scale"}),
            "binom": (dist.Binomial, {"n":"total_count", "p":"probs"}),
            "normal": dist.Normal,
            "halfnorm": HalfNormalTrans,
            "poisson": (dist.Poisson, {"mu": "rate"}),
        }
        
        self.simulation = simulation
        self.evaluator = self.parse_deterministic_model()
        self.prior = self.parse_model_priors()
        self.error_model = self.parse_error_model()

        # parse preprocessing
        if self.user_defined_preprocessing is not None:
            self.preprocessing = getattr(
                self.simulation.prob,
                self.user_defined_preprocessing
            )

        # parse the probability model
        if self.user_defined_probability_model is not None:
            self.inference_model = getattr(
                self.simulation.prob, 
                self.user_defined_probability_model
            )
        elif self.gaussian_base_distribution:
            self.inference_model = self.parse_probabilistic_model_with_gaussian_base()
        else:
            self.inference_model = self.parse_probabilistic_model()


        self.idata: az.InferenceData

    @property
    def plot_function(self) -> Optional[str]:
        return self.simulation.config.get(
            "inference", "plot_function", 
            fallback=None
        )

    @property
    def extra_vars(self):
        extra = self.simulation.config.getlist( 
            "inference", "extra_vars", fallback=[]
        )
        return extra if isinstance(extra, list) else [extra]
    

    @property
    def user_defined_probability_model(self):
        return self.simulation.config.get(
            "inference.numpyro", "user_defined_probability_model", fallback=None
        )
    
    @property
    def user_defined_preprocessing(self):
        return self.simulation.config.get(
            "inference.numpyro", "user_defined_preprocessing", fallback=None
        )

    @property
    def gaussian_base_distribution(self):
        flag = self.simulation.config.getint(
            "inference.numpyro", "gaussian_base_distribution", fallback=False
        )
        return bool(flag)
    
    @property
    def n_predictions(self):
        return self.simulation.config.getint(
            "inference", "n_predictions", fallback=100
        )
    

    @property
    def chains(self):
        return self.simulation.config.getint(
            "inference.numpyro", "chains", fallback=1
        )
    
    @property
    def draws(self):
        return self.simulation.config.getint(
            "inference.numpyro", "draws", fallback=1000
        )
    
    @property
    def warmup(self):
        return self.simulation.config.getint(
            "inference.numpyro", "warmup", fallback=self.draws
        )
    
    @property
    def thinning(self):
        return self.simulation.config.getint(
            "inference.numpyro", "thinning", fallback=1
        )

    @property
    def svi_iterations(self):
        return self.simulation.config.getint(
            "inference.numpyro", "svi_iterations", fallback=10_000
        )
    
    @property
    def svi_learning_rate(self):
        return self.simulation.config.getfloat(
            "inference.numpyro", "svi_learning_rate", fallback=0.0001
        )
    
    @property
    def kernel(self):
        return self.simulation.config.get(
            "inference.numpyro", "kernel", fallback="nuts"
        )
    

    @property
    def adapt_state_size(self):
        return self.simulation.config.getint(
            "inference.numpyro", "sa_adapt_state_size", fallback=None
        )

    @property
    def init_strategy(self):
        strategy = self.simulation.config.get(
            "inference.numpyro", "init_strategy", fallback="init_to_median"
        )

        return getattr(infer, strategy)

    def parse_deterministic_model(self) -> Callable:
        """Parses an evaluation function from the Simulation object, which 
        takes a single argument theta and defaults to passing no seed to the
        deterministic evaluator.

        Returns
        -------
        callable
            The evaluation function
        """
        def evaluator(theta, seed=None):
            evaluator = self.simulation.dispatch(theta=theta)
            evaluator(seed)
            return evaluator.Y
        
        return evaluator

    def model(self):
        pass


    def observation_parser(self) -> Tuple[Dict,Dict]:
        """Transform a xarray.Dataset into a dictionary of jnp.Arrays. Creates
        boolean arrays of masks for nan values (missing values are tagged False)

        Returns
        -------
        Tuple[Dict,Dict]
            Dictionaries of observations (data) and masks (missing values)
        """
        obs = self.simulation.observations \
            .transpose(*self.simulation.dimensions)
        data_vars = self.simulation.data_variables + self.extra_vars

        masks = {}
        observations = {}
        for d in data_vars:
            o = jnp.array(obs[d].values)
            m = jnp.logical_not(jnp.isnan(o))
            observations.update({d:o})
            masks.update({d:m})
        
        return observations, masks
    
    @staticmethod
    def parse_parameter(parname: str, prior: str, distribution_map: Dict[str,Tuple]) -> Tuple[str,dist.Distribution,Dict[str,Callable]]:

        distribution, cluttered_arguments = prior.split("(", 1)
        param_strings = cluttered_arguments.split(")", 1)[0].split(",")

        distribution_mapping = distribution_map[distribution]

        if not isinstance(distribution_mapping, tuple):
            distribution = distribution_mapping
            distribution_mapping = (distribution, {})
        
        assert len(distribution_mapping) == 2, (
            "distribution and parameter mapping must be "
            "a tuple of length 2."
        )

        distribution, parameter_mapping = distribution_mapping
        mapped_params = {}
        for parstr in param_strings:
            key, val = parstr.split("=")
            
            mapped_key = parameter_mapping.get(key, key)

            # if is_number(val):
            #     parsed_val = float(val)
            # else:
            parsed_val = generate_transform(expression_str=val)

            # parsed_val = float(val) if is_number(val) else val
            mapped_params.update({mapped_key:parsed_val})

        return parname, distribution, mapped_params

    def parse_model_priors(self):
        priors = {}
        for par in self.simulation.free_model_parameters:
            name, distribution, params = self.parse_parameter(
                parname=par.name, 
                prior=par.prior, 
                distribution_map=self.distribution_map
            )
            # parameterized_dist = distribution(**params)
            priors.update({
                name: {
                    "fn":distribution, 
                    "parameters": params
                }
            })
        return priors

    def parse_error_model(self):
        error_model = {}
        for data_var, error_distribution in self.simulation.error_model.items():
            name, distribution, parameters = self.parse_parameter(
                parname=data_var,
                prior=error_distribution,
                distribution_map=self.distribution_map
            )

                
            error_model.update({
                data_var: {
                    "fn": distribution, 
                    "parameters": parameters,
                    "error_dist": error_distribution
                }
            })
        return error_model

    def parse_probabilistic_model(self):
        EPS = self.EPS
        prior = self.prior.copy()
        error_model = self.error_model.copy()
        extra = {"EPS": EPS}

        def lookup(val, deterministic, prior_samples, observations):
            if val in deterministic:
                return deterministic[val]
            
            elif val in prior_samples:
                return prior_samples[val]
            
            elif val in observations:
                return observations[val]
            
            elif val in extra:
                return extra[val]

            else:
                return val

        def model(solver, obs, masks):
            # construct priors with numpyro.sample and sample during inference
            theta = {}
            for prior_name, prior_kwargs in prior.items():
                
                # apply transforms to priors. This will be handy when I use nested
                # parameters
                prior_distribution_parameters = {}
                for pri_par, pri_val in prior_kwargs["parameters"].items():
                    prior_trans_func = pri_val["transform"]
                    prior_trans_func_kwargs = {k: lookup(k, {}, theta, obs) for k in pri_val["args"]}
                    prior_distribution_parameters.update({
                        pri_par: prior_trans_func(**prior_trans_func_kwargs)
                    })

                theta_i = numpyro.sample(
                    name=prior_name,
                    fn=prior_kwargs["fn"](**prior_distribution_parameters)
                )

                theta.update({prior_name: theta_i})
            
            # calculate deterministic simulation with parameter samples
            sim_results = solver(theta=theta)

            # store data_variables as deterministic model output
            for deterministic_name, deterministic_value in sim_results.items():
                _ = numpyro.deterministic(
                    name=deterministic_name, 
                    value=deterministic_value
                )

            for error_model_name, error_model_kwargs in error_model.items():
                error_distribution = error_model_kwargs["fn"]
                error_distribution_kwargs = {}
                for errdist_par, errdist_val in error_model_kwargs["parameters"].items():
                    errdist_trans_func = errdist_val["transform"]
                    errdist_trans_func_kwargs = {
                        k: lookup(k, sim_results, theta, obs) for k in errdist_val["args"]
                    }
                    error_distribution_kwargs.update({
                        errdist_par: errdist_trans_func(**errdist_trans_func_kwargs)
                    })

                _ = numpyro.sample(
                    name=error_model_name + "_obs",
                    fn=error_distribution(**error_distribution_kwargs).mask(masks[error_model_name]),
                    obs=obs[error_model_name]
                )


            # TODO: How to add EPS
            # DONE: add biomial n --> I think this is already done
            # numpyro.sample("cext_obs", LogNormalTrans(loc=cext + EPS, scale=sigma_cext).mask(masks["cext"]), obs=obs["cext"] + EPS)
            # numpyro.sample("cint_obs", LogNormalTrans(loc=cint + EPS, scale=sigma_cint).mask(masks["cint"]), obs=obs["cint"] + EPS)
            # numpyro.sample("nrf2_obs", LogNormalTrans(loc=nrf2 + EPS, scale=sigma_nrf2).mask(masks["nrf2"]), obs=obs["nrf2"] + EPS)
            # numpyro.sample("lethality_obs", dist.Binomial(probs=leth, total_count=obs["nzfe"]).mask(masks["lethality"]), obs=obs["lethality"])
        return model

    def parse_probabilistic_model_with_gaussian_base(self):
        EPS = self.EPS
        prior = self.prior.copy()
        error_model = self.error_model.copy()
        extra = {"EPS": EPS}


        def lookup(val, deterministic, prior_samples, observations):
            if val in deterministic:
                return deterministic[val]
            
            elif val in prior_samples:
                return prior_samples[val]
            
            elif val in observations:
                return observations[val]
            
            elif val in extra:
                return extra[val]

            else:
                return val

        def model(solver, obs, masks):
            theta = {}
            theta_base = {}
            for prior_name, prior_kwargs in prior.items():

                # apply transforms to priors. This will be handy when I use nested
                # parameters
                prior_distribution_parameters = {}
                for pri_par, pri_val in prior_kwargs["parameters"].items():
                    prior_trans_func = pri_val["transform"]
                    # TODO could be replaced by utils.config.lookup and lookup args
                    prior_trans_func_kwargs = {k: lookup(k, {}, theta, obs) for k in pri_val["args"]}
                    prior_distribution_parameters.update({
                        pri_par: prior_trans_func(**prior_trans_func_kwargs)
                    })

                # parameterize the prior distribution and extract transforms
                dist = prior_kwargs["fn"](**prior_distribution_parameters)
                
                try:
                    transforms = getattr(dist, "transforms")
                except:
                    raise RuntimeError(
                        "The specified distribution had no transforms. If setting "
                        "the option 'inference.numpyro.gaussian_base_distribution = 1', "
                        "you are only allowed to use parameter distribution, which can "
                        "be specified as transformed normal distributions. "
                        "Currently only 'lognorm' and 'halfnorm' are implemented. "
                        "You can use the numypro.distributions.TransformedDistribution "
                        "API to specify additional distributions with transforms."
                        "And pass them to the inferer by updating the distribution map: "
                        "sim.inferer.distribution_map.update({'newdist': your_new_distribution})"
                    )

                # sample from a random normal distribution
                theta_base_i = numpyro.sample(
                    name=f"{prior_name}_normal_base",
                    fn=Normal(loc=0, scale=1),
                    sample_shape=dist.shape()                    
                )

                # apply the transforms 
                theta_i = numpyro.deterministic(
                    name=prior_name,
                    value=transform(transforms=transforms, x=theta_base_i)
                )

                theta_base.update({prior_name: theta_base_i})
                theta.update({prior_name: theta_i})

            # calculate deterministic simulation with parameter samples
            sim_results = solver(theta=theta)

            # store data_variables as deterministic model output
            for deterministic_name, deterministic_value in sim_results.items():
                _ = numpyro.deterministic(
                    name=deterministic_name, 
                    value=deterministic_value
                )

            for error_model_name, error_model_kwargs in error_model.items():
                error_distribution = error_model_kwargs["fn"]
                error_distribution_kwargs = {}
                for errdist_par, errdist_val in error_model_kwargs["parameters"].items():
                    errdist_trans_func = errdist_val["transform"]
                    errdist_trans_func_kwargs = {
                        k: lookup(k, sim_results, theta, obs) for k in errdist_val["args"]
                    }
                    error_distribution_kwargs.update({
                        errdist_par: errdist_trans_func(**errdist_trans_func_kwargs)
                    })

                _ = numpyro.sample(
                    name=error_model_name + "_obs",
                    fn=error_distribution(**error_distribution_kwargs).mask(masks[error_model_name]),
                    obs=obs[error_model_name]
                )


            # TODO: How to add EPS
            # DONE: add biomial n --> I think this is already done
            # numpyro.sample("cext_obs", LogNormalTrans(loc=cext + EPS, scale=sigma_cext).mask(masks["cext"]), obs=obs["cext"] + EPS)
            # numpyro.sample("cint_obs", LogNormalTrans(loc=cint + EPS, scale=sigma_cint).mask(masks["cint"]), obs=obs["cint"] + EPS)
            # numpyro.sample("nrf2_obs", LogNormalTrans(loc=nrf2 + EPS, scale=sigma_nrf2).mask(masks["nrf2"]), obs=obs["nrf2"] + EPS)
            # numpyro.sample("lethality_obs", dist.Binomial(probs=leth, total_count=obs["nzfe"]).mask(masks["lethality"]), obs=obs["lethality"])
        return model


    @staticmethod
    def preprocessing(**kwargs):
        return kwargs

    def run(self, print_debug=True):
        # set parameters of JAX and numpyro
        # jax.config.update("jax_enable_x64", True)

        # generate random keys
        key = jax.random.PRNGKey(self.simulation.seed)
        key, *subkeys = jax.random.split(key, 20)
        keys = iter(subkeys)

        # parse observations and masks for missing data
        obs, masks = self.observation_parser()

        model_kwargs = self.preprocessing(
            obs=obs, 
            masks=masks,
        )

        # prepare model and print information about shapes
        model = partial(
            self.inference_model, 
            solver=self.evaluator, 
            **model_kwargs
        )    

        try:
            import graphviz
            graph = numpyro.render_model(model)
            graph.render(
                filename=f"{self.simulation.output_path}/probability_model",
                view=False, cleanup=True, format="png"
            ) 
        except graphviz.backend.ExecutableNotFound:
            warnings.warn(
                "Model is not rendered, because the graphviz executable is "
                "not found. Try search for 'graphviz executables not found'"
                "and the used OS. This should be an easy fix :-)"
            )


        if print_debug:
            with numpyro.handlers.seed(rng_seed=1):
                trace = numpyro.handlers.trace(model).get_trace()
            print(numpyro.util.format_shapes(trace))
            
        # run inference
        if self.kernel.lower() == "sa" or self.kernel.lower() == "nuts":
            sampler, mcmc = self.run_mcmc(
                model=model,
                keys=keys,
                kernel=self.kernel.lower()
            )

            # create arviz idata
            self.idata = az.from_numpyro(
                mcmc, 
                dims=self.posterior_data_structure,
                coords=self.posterior_coordinates,
            )

        elif self.kernel.lower() == "svi" or self.kernel.lower() == "map":
            if not self.gaussian_base_distribution:
                raise RuntimeError(
                    "SVI is only supported if parameter distributions can be "
                    "re-parameterized as gaussians. Please set "
                    "inference.numpyro.gaussian_base_distribution = 1 "
                    "and if needed use distributions from the loc-scale family "
                    "to specify the model parameters."
                )
            
            svi, guide, svi_result = self.run_svi(
                model=model,
                keys=keys,
                kernel=self.kernel.lower(),
                learning_rate=self.svi_learning_rate,
                iterations=self.svi_iterations,
            )

            # plot loss curve
            fig, ax = plt.subplots(1, 1)
            ax.plot(svi_result.losses)
            ax.set_yscale("log")
            ax.set_ylabel("Loss")
            ax.set_xlabel("Iteration")
            fig.savefig(f"{self.simulation.output_path}/svi_loss_curve.png")

            # save idata and print summary
            draws = 1 if self.kernel.lower() == "map" else self.draws
            self.idata = self.svi_posterior(
                svi_result, model, guide, next(keys), draws
            )
            print(az.summary(self.idata))

        else:
            raise NotImplementedError(
                f"Kernel {self.kernel} is not implemented. "
                "Use one of nuts, sa, svi, map"
            )
        
    def run_mcmc(self, model, keys, kernel):
        if kernel == "sa":
            sampler = infer.SA(
                model=model,
                dense_mass=True,
                adapt_state_size=self.adapt_state_size,
                init_strategy=self.init_strategy,
            )

        elif kernel == "nuts":
            sampler = infer.NUTS(
                model, 
                dense_mass=True, 
                step_size=0.01,
                adapt_mass_matrix=True,
                adapt_step_size=True,
                max_tree_depth=10,
                target_accept_prob=0.8,
                init_strategy=self.init_strategy
            )
        else:
            raise NotImplementedError(
                f"MCMC kernel {kernel} not implemented. Use one of 'sa', 'nuts'"
            )


        mcmc = infer.MCMC(
            sampler=sampler,
            num_warmup=self.warmup,
            num_samples=self.draws * self.thinning,
            num_chains=self.chains,
            thinning=self.thinning,
            progress_bar=True,
        )
    
        # run inference
        mcmc.run(next(keys))
        mcmc.print_summary()

        return sampler, mcmc

    @staticmethod
    def run_svi(model, keys, learning_rate, iterations, kernel):

        init_fn = partial(infer.init_to_uniform, radius=1)
        if kernel == "svi":
            guide = infer.autoguide.AutoMultivariateNormal(model, init_loc_fn=init_fn)
        elif kernel == "map":
            guide = numpyro.infer.autoguide.AutoDelta(model, init_loc_fn=init_fn)
        else:
            raise NotImplementedError(
                f"SVI kernel {kernel} is not implemented. "
                "Use one of 'map', 'svi'"
            )

        optimizer = numpyro.optim.ClippedAdam(step_size=learning_rate, clip_norm=10)
        svi = infer.SVI(model=model, guide=guide, optim=optimizer, loss=infer.Trace_ELBO())
        svi_result = svi.run(next(keys), iterations, stable_update=True)

        if kernel == "svi":
            cov = svi_result.params['auto_scale_tril'].dot(
                svi_result.params['auto_scale_tril'].T
            )
            median = guide.median(svi_result.params)

        return svi, guide, svi_result

    @property
    def posterior(self):
        return self.idata.posterior


    @property
    def posterior_data_structure(self):
        data_structure = self.simulation.data_structure.copy()
        data_structure_loglik = {f"{dv}_obs": dims for dv, dims in data_structure.items()}
        data_structure.update(data_structure_loglik)
        return data_structure
    
    @property
    def posterior_coordinates(self):
        posterior_coords = self.simulation.coordinates.copy()
        posterior_coords.update({
            "draw": list(range(self.draws)), 
            "chain": list(range(self.chains))
        })
        return posterior_coords
    

    def create_log_likelihood(self, seed=1):
        key = jax.random.PRNGKey(seed)
        obs, masks = self.observation_parser()

        model_kwargs = self.preprocessing(
            obs=obs, 
            masks=masks,
        )
        
        # prepare model
        model = partial(
            self.inference_model, 
            solver=self.evaluator, 
            **model_kwargs
        )    

        seeded_model = numpyro.handlers.seed(model, key)
   
        def log_density(theta, return_type="summed-by-site", check=True):
            """Log density relies heavily on the substitute utility
            
            The log density is synonymous for log-likelihood (it is the log 
            probability density of the model). 

            The general method is actually quite simple. Values of all SAMPLE
            sites are replaced according to the key: value pairs in `theta`.

            Then the model is calculated and the trace is obtained. Everything
            else is then just post-processing of the sites. Here the log_prob
            function of the sites in the trace are used and the values of the
            sites are inserted. 

            Note that the log-density can randomly fluctuate, if not all
            sites are replaced.

            Note that the data-loglik can be used to calculate a maximum-likelihood
            estimate. Because it is independent of the prior

            Parameters
            ----------

            theta : Dict
                Dictionary of priors (sites) which should be deterministically 
                fixed (substituted).

            return_type : str
                The information which should be returned. With increasing level
                of computation:
                
                joint-log-likelihood: returns a single value, the entire log
                    likelihood of the model, given the values in theta
                full: joint-log, loglik-prior of each site and value, 
                    loglik-data of each site and value 
                summed-by-site: joint-loglik, loglik-prior of sites, 
                    loglik-data of sites
                summed-by-prior-data:
                    joint-loglik, prior-loglik, data-loglik

            """
            
            # TODO: For speeding this up, get inspired by
            # https://num.pyro.ai/en/stable/_modules/numpyro/infer/util.html#log_likelihood

        
            joint_log_density, trace = numpyro.infer.util.log_density(
                model=seeded_model,
                model_args=(),
                model_kwargs={},
                params=theta
            )

            joint_log_density = float(joint_log_density)

            if check:
                sites_in_theta = {
                    name: True 
                    if name in theta
                    else False
                    for name, site in {
                        name: site for name, site in trace.items() 
                        if site["type"] == "sample" 
                        and not site["is_observed"]
                    }.items()
                }

                if not all(list(sites_in_theta.values())):
                    missing_sites = [name for name, in_theta in sites_in_theta.items() if not in_theta]
                    warnings.warn(
                        f"Sites: {missing_sites} were not specified in theta. "
                        "Log-likelihood will not be fully defined by theta."
                        "Results should be independent of the given seed"
                    )


            if return_type == "joint-log-likelihood":
                return joint_log_density
            
            
            prior_loglik = {
                name: site["fn"].log_prob(site["value"])
                for name, site in trace.items()
                if site["type"] == "sample" and not site["is_observed"]
            }

            data_loglik = {
                name: site["fn"].log_prob(site["value"])
                for name, site in trace.items()
                if site["type"] == "sample" and site["is_observed"]
            }

            if return_type == "full":
                return joint_log_density, prior_loglik, data_loglik

            prior_loglik_sum = {
                key: np.sum(value) for key, value in prior_loglik.items()
            }

            data_loglik_sum = {
                key: np.sum(value) for key, value in data_loglik.items()
            }

            if return_type == "summed-by-site":
                return joint_log_density, prior_loglik_sum, data_loglik_sum

            prior_loglik_total = np.sum(list(prior_loglik_sum.values()))
            data_loglik_total = np.sum(list(data_loglik_sum.values()))
            
            if return_type == "summed-by-prior-data":
                return joint_log_density, prior_loglik_total, data_loglik_total
            
            raise NotImplementedError(f"return_type flag: {return_type} is not implemented")


        return log_density

    @lru_cache
    def prior_predictions(self, n=100, seed=1):
        key = jax.random.PRNGKey(seed)
        obs, masks = self.observation_parser()

        model_kwargs = self.preprocessing(
            obs=obs, 
            masks=masks,
        )
        
        # prepare model
        model = partial(
            self.inference_model, 
            solver=self.evaluator, 
            **model_kwargs
        )    
   
        prior_predictive = Predictive(
            model, num_samples=n, batch_ndims=2
        )
        prior_predictions = prior_predictive(key)

        loglik = numpyro.infer.log_likelihood(
            model=model, 
            posterior_samples=prior_predictions, 
            batch_ndims=2, 
            obs=obs, 
            masks=masks
        )

        preds = self.simulation.data_variables
        preds_obs = [f"{d}_obs" for d in self.simulation.data_variables]
        priors = list(self.simulation.model_parameter_dict.keys())
        # TODO: This causes priors to be dropped that are not part of the 
        #       config file. Makes interactive model development more difficult
        #       This should be derived from the model directly.
        posterior_coords = self.posterior_coordinates
        posterior_coords["draw"] = list(range(n))
        data_structure = self.posterior_data_structure
        
        idata = az.from_dict(
            observed_data=obs,
            prior={k: v for k, v in prior_predictions.items() if k in priors},
            prior_predictive={k: v for k, v in prior_predictions.items() if k in preds+preds_obs},
            log_likelihood=loglik,
            dims=data_structure,
            coords=posterior_coords,
        )

        return idata
        # self.idata.to_netcdf(f"{self.simulation.output_path}/numpyro_prior_predictions.nc")
    

    def svi_posterior(self, svi_result, model, guide, key, n=1000, only_parameters=False):
        key, *subkeys = jax.random.split(key, 4)
        keys = iter(subkeys)

        obs, masks = self.observation_parser()

        params = svi_result.params

        # predictive = Predictive(
        #     model, guide=guide, params=params, 
        #     num_samples=n, batch_ndims=2
        # )
        # samples = predictive(next(keys))    

        predictive = Predictive(
            guide, params=params, 
            num_samples=n, batch_ndims=2
        )
        posterior_samples = predictive(next(keys))

        predictive = Predictive(
            model, posterior_samples, params=params, 
            num_samples=n, batch_ndims=2
        )
        posterior_predictions = predictive(next(keys))


        loglik = numpyro.infer.log_likelihood(
            model=model, 
            posterior_samples=posterior_samples, 
            batch_ndims=2, 
        )

        preds = self.simulation.data_variables
        preds_obs = [f"{d}_obs" for d in self.simulation.data_variables]
        priors = list(self.simulation.model_parameter_dict.keys())
        posterior_coords = self.posterior_coordinates
        posterior_coords["draw"] = list(range(n))
        data_structure = self.posterior_data_structure

        # TODO add prior data structure. Address this when a proper coordinate
        # dimensionality backend is implemented (config module)
        
        # TODO add option to only return parameters (posterior) without
        # predictions [SHOULD BE EXPOSED ON STORING THE POSTERIOR]. 
        # Do keep calculating the predictions. They can be used
        # for quick and dirty AND STANDARDIZED diagnoses

        idata = az.from_dict(
            observed_data=obs,
            posterior={k: v for k, v in posterior_predictions.items() if k in priors},
            posterior_predictive={k: v for k, v in posterior_predictions.items() if k in preds},
            log_likelihood=loglik,
            dims=data_structure,
            coords=posterior_coords,
        )

        return idata
    

    @staticmethod
    def get_dict(group: xr.Dataset):
        data_dict = group.to_dict()["data_vars"]
        return {k: np.array(val["data"]) for k, val in data_dict.items()}


    @lru_cache
    def posterior_predictions(self, n: Optional[int]=None, seed=1):
        # TODO: It may be necessary that the coordinates should be passed as 
        # constant data. Because if the model is compiled with them once, 
        # according to the design philosophy of JAX, the model will not 
        # be evaluated again. But considering that the jitted functions do take
        # coordinates as an input argument, maybe I'm okay. This should be
        # tested.
        posterior = self.idata.posterior
        n_samples = posterior.sizes["chain"] * posterior.sizes["draw"]
    
        if n is not None:
            key = jax.random.PRNGKey(seed)

            n_draws = int(n / posterior.sizes["chain"])
            # the same selection of draws will be applied to all chains. This
            # any other form will result in an array, where a lot of nans 
            # are present of the size n * chains, while we want size n
            selection = jax.random.choice(
                key=key, 
                a=posterior.draw.values, 
                replace=False, 
                shape=(n_draws, )
            )
            posterior = posterior.isel(draw=selection)


        preds = []
        with tqdm(
            total=posterior.sizes["chain"] * posterior.sizes["draw"],
            desc="Posterior predictions"
        ) as pbar:
            for chain in posterior.chain:
                for draw in posterior.draw:
                    theta = posterior.sel(draw=draw, chain=chain)
                    evaluator = self.simulation.dispatch(theta=self.get_dict(theta))
                    evaluator()
                    ds = evaluator.results

                    ds = ds.assign_coords({"chain": chain, "draw": draw})
                    ds = ds.expand_dims(("chain", "draw"))
                    preds.append(ds)
                    pbar.update(1)

        # key = jax.random.PRNGKey(seed)
        # model = partial(self.model, solver=self.evaluator)    
        # predict = numpyro.infer.Predictive(model, posterior_samples=posterior, batch_ndims=2)
        # predict(key, obs=obs, masks=masks)
        

        return xr.combine_by_coords(preds)

    def store_results(self, output=None):
        if output is not None:
            self.idata.to_netcdf(output)
        else:
            self.idata.to_netcdf(f"{self.simulation.output_path}/numpyro_posterior.nc")

    def load_results(self, file="numpyro_posterior.nc"):
        self.idata = az.from_netcdf(f"{self.simulation.output_path}/{file}")


    def plot(self):
        # TODO: combine prior_predictions and posterior predictions
        if hasattr(self.idata, "posterior"):
            axes = az.plot_trace(
                self.idata,
                var_names=self.simulation.model_parameter_names
            )
            fig = plt.gcf()
            fig.savefig(f"{self.simulation.output_path}/trace.png")
            plt.close()
            axes = az.plot_pair(
                self.idata, 
                divergences=True, 
                var_names=self.simulation.model_parameter_names
            )
            fig = plt.gcf()
            fig.savefig(f"{self.simulation.output_path}/pairs_posterior.png")
            plt.close()

        if hasattr(self.idata, "prior_predictive"):
            self.plot_prior_predictive()

        plot_func = getattr(self.simulation, self.plot_function)
        plot_func()

    def plot_prior_predictions(
            self, data_variable: str, x_dim: str, ax=None, subset={}, 
            n=None, seed=None
        ):
        if n is None:
            n = self.n_predictions

        if seed is None:
            seed = self.simulation.seed
        
        idata = self.prior_predictions(
            n=n, 
            # seed only controls the parameters samples drawn from posterior
            seed=seed
        )
        plot_func = getattr(self.simulation, self.plot_function)
        
        ax = self.plot_predictions(
            observations=self.simulation.observations,
            predictions=idata.prior_predictive,
            data_variable=data_variable,
            x_dim=x_dim,
            ax=ax,
            subset=subset,
            mode="draws"
        )

        return ax

    def plot_posterior_predictions(
            self, data_variable: str, x_dim: str, ax=None, subset={},
            n=None, seed=None
        ):
        # TODO: This method should be trashed. It is not really useful
        if n is None:
            n = self.n_predictions
        
        if seed is None:
            seed = self.simulation.seed
        
        predictions = self.posterior_predictions(
            n=n, 
            # seed only controls the parameters samples drawn from posterior
            seed=seed
        )
        
        ax = self.plot_predictions(
            observations=self.simulation.observations,
            predictions=predictions,
            data_variable=data_variable,
            x_dim=x_dim,
            ax=ax,
            subset=subset,
        )

        return ax



    def plot_predictions(
            self, 
            observations,
            predictions,
            data_variable: str, 
            x_dim: str, 
            ax=None, 
            subset={},
            mode="mean+hdi",
            plot_options={"obs": {}, "pred_mean": {}, "pred_draws": {}, "pred_hdi": {}},
        ):
        # filter subset coordinates present in data_variable
        subset = {k: v for k, v in subset.items() if k in observations.coords}
        
        # select subset
        obs = observations.sel(subset)[data_variable]
        preds = predictions.sel(subset)[f"{data_variable}"]
        
        # stack all dims that are not in the time dimension
        if len(obs.dims) == 1:
            # add a dummy batch dimension
            obs = obs.expand_dims("batch")
            obs = obs.assign_coords(batch=[0])

            preds = preds.expand_dims("batch")
            preds = preds.assign_coords(batch=[0])


        stack_dims = [d for d in obs.dims if d != x_dim]
        obs = obs.stack(i=stack_dims)
        preds = preds.stack(i=stack_dims)
        N = len(obs.coords["i"])
            
        hdi = az.hdi(preds, .95)[f"{data_variable}"]

        if ax is None:
            ax = plt.subplot(111)
        
        y_mean = preds.mean(dim=("chain", "draw"))

        for i in obs.i:
            if obs.sel(i=i).isnull().all():
                # skip plotting combinations, where all values are NaN
                continue
            
            if mode == "mean+hdi":
                kwargs_hdi = dict(color="black", alpha=0.1)
                kwargs_hdi.update(plot_options.get("pred_hdi", {}))
                ax.fill_between(
                    preds[x_dim].values, *hdi.sel(i=i).values.T, 
                    **kwargs_hdi
                )

                kwargs_mean = dict(color="black", lw=1, alpha=max(1/N, 0.05))
                kwargs_mean.update(plot_options.get("pred_mean", {}))
                ax.plot(
                    preds[x_dim].values, y_mean.sel(i=i).values, 
                    **kwargs_mean
                )
            elif mode == "draws":
                kwargs_draws = dict(color="black", lw=0.5, alpha=max(1/N, 0.05))
                kwargs_draws.update(plot_options.get("pred_draws", {}))
                ys = preds.sel(i=i).stack(sample=("chain", "draw"))
                ax.plot(
                    preds[x_dim].values, ys.values, 
                    **kwargs_draws
                )
            else:
                raise NotImplementedError(
                    f"Mode '{mode}' not implemented. "
                    "Choose 'mean+hdi' or 'draws'."
                )

            kwargs_obs = dict(marker="o", ls="", ms=3, color="tab:blue")
            kwargs_obs.update(plot_options.get("obs", {}))
            ax.plot(
                obs[x_dim].values, obs.sel(i=i).values, 
                **kwargs_obs
            )
        
        ax.set_ylabel(data_variable)
        ax.set_xlabel(x_dim)

        return ax

    # This is a separate script!    
    def combine_chains(self, chain_location="chains", drop_extra_vars=[], cluster_deviation="std"):
        """Combine chains if chains were computed in a fully parallelized manner
        (on different machines, jobs, etc.). 

        In addition, the method drops all data variables and *_norm priors 
        (i.e. helper priors with a normal base). This is done, in order to
        create slim data objects for storage.

        Parameters
        ----------
        chain_location : str, optional
            location of the chains, relative to the simulation.output_path, this
            parameter is simulteneously the string appended to the saved 
            posterior. By default "chains"
        drop_extra_vars : List, optional
            any additional variables to drop from the posterior
        """
        sim = self.simulation
        pseudo_chains = glob.glob(
            f"{sim.output_path}/{chain_location}/*/numpyro_posterior.nc"
        )

        # just be aware that in the case of MAP this is not an acutal posterior.
        # But it can behave like one with multiple chains (1 for each start)
        idata = az.from_netcdf(pseudo_chains[0])
        posterior = self.drop_vars_from_posterior(idata.posterior, drop_extra_vars)
        log_likelihood = idata.log_likelihood

        # iterate over the posterior files with a progress bar (depending on the
        # size and number of posteriors this op needs time and memory)
        tqdm_iterator = tqdm(
            enumerate(pseudo_chains[1:], start=1), 
            total=len(pseudo_chains)-1,
            desc="Concatenating posteriors"
        )
        for i, f in tqdm_iterator:
            idata = az.from_netcdf(f)
            ccord = {"chain": np.array([i])}
            
            # add chain coordinate to posterior and likelihood
            idata.posterior = self.drop_vars_from_posterior(idata.posterior, drop_extra_vars)
            idata.posterior = idata.posterior.assign_coords(ccord)
            idata.log_likelihood = idata.log_likelihood.assign_coords(ccord)

            # concatenate chains
            posterior = xr.concat([posterior, idata.posterior], dim="chain")
            log_likelihood = xr.concat(
                [log_likelihood, idata.log_likelihood], 
                dim="chain"
            )

        posterior = rename_extra_dims(
            posterior, 
            new_dim="substance", 
            new_coords=sim.observations.attrs["substance"]
        )

        # store mutlichain inferencedata to the main output directory
        # this is also a slim posterior that only contains the necessary information
        # posterior and likelihood and has therefore a small file size
        idata_multichain = az.InferenceData(
            posterior=posterior, 
            log_likelihood=log_likelihood,
            observed_data=idata.observed_data,
        )

        idata_multichain = add_cluster_coordinates(idata_multichain, cluster_deviation)
        print("Clusters:", idata_multichain.posterior.cluster)

        return idata_multichain            

    def drop_vars_from_posterior(self, posterior, drop_extra_vars):
        """drops extra variables if they are included in the posterior
        """
        drop_vars = [k for k in list(posterior.data_vars.keys()) if "_norm" in k]
        drop_vars = drop_vars + self.simulation.data_variables + drop_extra_vars
        drop_vars = [v for v in drop_vars if v in posterior]
        drop_coords = [c for c in list(posterior.coords.keys()) if c.split("_dim_")[0] in drop_vars]

        posterior = posterior.drop(drop_vars)
        posterior = posterior.drop(drop_coords)

        return posterior


import subprocess
import xarray as xr
import numpy as np
from click.testing import CliRunner
from matplotlib import pyplot as plt

from pymob.utils.store_file import prepare_casestudy

def create_simulation():
    config = prepare_casestudy(
        case_study=("test_case_study", "test_scenario"),
        config_file="settings.cfg"
    )
    from case_studies.test_case_study.sim import Simulation
    return Simulation(config=config)

def test_scripting_api_pyabc():
    sim = create_simulation()
    sim.set_inferer(backend="pyabc")
    sim.inferer.run()
    sim.inferer.store_results()
    sim.inferer.load_results()
    
    posterior_mean = sim.inferer.idata.posterior.mean(("chain", "draw"))
    true_parameters = sim.model_parameter_dict
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_mean.to_dataarray().values,
        np.array(list(true_parameters.values())),
        rtol=5e-2, atol=1e-5
    )


def test_inference_evaluation():
    sim = create_simulation()
    sim.set_inferer(backend="pyabc")

    sim.inferer.load_results()
    fig = sim.inferer.plot_chains()
    fig.savefig(sim.output_path + "/pyabc_chains.png")

    # posterior predictions
    for data_var in sim.data_variables:
        ax = sim.inferer.plot_predictions(
            data_variable=data_var, 
            x_dim="time"
        )
        fig = ax.get_figure()

        fig.savefig(f"{sim.output_path}/pyabc_posterior_predictions_{data_var}.png")
        plt.close()

def test_commandline_API_redis():
    from pymob.infer import main
    runner = CliRunner()
    
    args = "--case_study=test_case_study --scenario=test_scenario"
    result = runner.invoke(main, args.split(" "))

def test_commandline_API_infer():
    from pymob.infer import main
    runner = CliRunner()
    
    args = "--case_study=test_case_study --scenario=test_scenario"
    result = runner.invoke(main, args.split(" "))

if __name__ == "__main__":
    import sys
    import os
    sys.path.append(os.getcwd())


from typing import Callable, Dict, List, Optional
import inspect
import xarray as xr
import numpy as np
from pymob.sim.solvetools import mappar

def create_dataset_from_numpy(Y, Y_names, coordinates):
    DeprecationWarning(
        "This method will be discontinued in future relases. "
        "Use 'create_dataset_from_dict' instead and return a dictionary from "
        "the solver or post processing respectively. The needed variable names "
        "to create the dictionary can be obtained from the data_variables "
        "argument in the solver signature. "
    )
    n_vars = Y.shape[-1]
    n_dims = len(Y.shape)
    assert n_vars == len(Y_names), (
        "The number of datasets must be the same as the specified number"
        "of data variables declared in the `settings.cfg` file."
    )

    # transpose Y to put the variable dimension first, then add the
    # remaining dimensions in order
    Y_transposed = Y.transpose((n_dims - 1, *range(n_dims - 1)))

    data_arrays = []
    for y, y_name in zip(Y_transposed, Y_names):
        da = xr.DataArray(y, coords=coordinates, name=y_name)
        data_arrays.append(da)

    dataset = xr.merge(data_arrays)

    return dataset

def create_dataset_from_dict(Y: dict, data_structure, coordinates):
    arrays = {}
    for k, v in Y.items():
        dims = data_structure.get(k, tuple(coordinates.keys()))
        coords = {d: coordinates[d] for d in dims}
        da = xr.DataArray(v, coords=coords, dims=dims)
        arrays.update({k: da})

    return xr.Dataset(arrays)

class Evaluator:
    """The Evaluator is an instance to evaluate a model. It's purpose is primarily
    to create objects that can be spawned and evaluated in parallel and can 
    individually track the results of a simulation or a parameter inference
    process. If needed the evaluations can be tracked and results can later
    be collected.
    
    Seed may not be set as a property, because this should be something passed
    through
    """
    result: xr.Dataset

    def __init__(
            self,
            model: Callable,
            solver: Callable,
            parameters: Dict,
            dimensions: List,
            n_ode_states: int,
            var_dim_mapper: List,
            data_structure: Dict,
            coordinates: Dict,
            data_variables: List,
            stochastic: bool,
            indices: Optional[Dict] = {},
            post_processing: Optional[Callable] = None,
            **kwargs
        ) -> None:
        """_summary_

        Parameters
        ----------
        model : Callable
            the ODE model to be solved by the evaluator
        solver : Callable
            a function to solve the ODE model with
        parameters : Dict
            A dictionary of model and post_processing parameters. Do not have
            to be in any particular order
        dimensions : List
            A list of the dimensions of the simulations
        n_ode_states : int
            The number of ODE states tracked
        var_dim_mapper : List
            A list of variables and their associated dimensions. This is relevant
            for simulations, where not all data variables have the same dimensional
            structure
        data_structure : Dict
            Similar to the var_dim_mapper, but additionally contains the coordinates
        coordinates : Dict
            The coordinates of each dimension in a dict
        data_variables : List
            The data variables of the simulation
        stochastic : bool
            Whether the model is a stochastic or a deterministic model
        indices : Optional[Dict], optional
            Indices, which should be used to map potentially nested parameters
            to a flat array for batch processing the simulations, by default {}
        post_processing : Optional[Callable], optional
            A function that takes a dictionary of simulation results and 
            parameters as an input and adds new variables to the results, 
            by default None, meaning that no post processing of the ODE solution
            is performed
        """
        
        self.model = model
        self.parameters = parameters
        self.dimensions = dimensions
        self.n_ode_states = n_ode_states
        self.var_dim_mapper = var_dim_mapper
        self.data_structure = data_structure
        self.data_variables = data_variables
        self.coordinates = coordinates
        self.is_stochastic = stochastic
        self.indices = indices
        
        if post_processing is None:
            self.post_processing = lambda results, time: results
        else: 
            self.post_processing = post_processing
                
        # set additional arguments of evaluator
        _ = [setattr(self, key, val) for key, val in kwargs.items()]

        self._signature = {}

        if isinstance(solver, type):
            self._solver = solver()
        else:
            self._solver = solver

        self.get_call_signature()



    def get_call_signature(self):
        signature = inspect.signature(self._solver)
        model_args = list(signature.parameters.keys())

        for a in model_args:
            if a not in self.allowed_model_signature_arguments:
                raise ValueError(
                    f"'{a}' in model signature is not an attribute of the Evaluator. "
                    f"Use one of {self.allowed_model_signature_arguments}, "
                    f"or set as evaluator_kwargs in the call to "
                    "'SimulationBase.dispatch'" 
                )
            
            # add argument to signature for call to model
            if a != "seed":
                self._signature.update({a: getattr(self, a)})
        
    
    @property
    def allowed_model_signature_arguments(self):
        return [a for a in self.__dict__.keys() if a[0] != "_"] + ["seed"]

    def __call__(self, seed=None):
        if seed is not None:
            self._signature.update({"seed": seed})

        Y_ = self._solver(**self._signature)
        
        # TODO: Consider which elements may be abstracted from the solve methods 
        # implemented in mod.py below is an unsuccessful approach
        # params = self._signature["parameters"]["parameters"]
        # time = self._signature["coordinates"]["time"]
        
        # s_dim, s_idx = self._signature["indices"]["substance"]
        # pp_args = mappar(self.post_processing, params, exclude=["t", "results"])
        # pp_args = [np.array(a, ndmin=1)[s_idx] for a in pp_args]
        # Y_ = self.post_processing(Y_, time, *pp_args)
        self.Y = Y_

    @property
    def dimensionality(self):
        return {key: len(values) for key, values in self.coordinates.items()}

    @property
    def results(self):
        if isinstance(self.Y, dict):
            return create_dataset_from_dict(
                Y=self.Y, 
                coordinates=self.coordinates,
                data_structure=self.data_structure,
            )
        elif isinstance(self.Y, np.ndarray):
            return create_dataset_from_numpy(
                Y=self.Y,
                Y_names=self.data_variables,
                coordinates=self.coordinates,
            )
        else:
            raise NotImplementedError(
                "Results returned by the solver must be of type Dict or np.ndarray."
            )
    